#include "sprites.h"
#include "config.h"
#include "global.h"
#include <cstdlib>
#include <stdexcept>


/*
 *  Trajectory
 */

Trajectory::Trajectory(const std::vector<RCoord> & arrival,
                       bool mirrored,
                       const Coord & base_cruise,
                       Coord delta_cruise)
: stage_       (arriving)
, arrival_     (arrival)
, arrival_it_  (arrival.begin())
, mirrored_    (mirrored)
, base_cruise_ (base_cruise)
, delta_cruise_(delta_cruise)
{
	if (arrival.empty()) {
		throw std::runtime_error ("Empty arrival trajectory");
	}
	pos_ = Coord(mirrored ? arrival[0].XMirror() : arrival[0]);
}


void Trajectory::NextPosition(int speed)
{
	if (speed == 0) return;

	switch (stage_) {
	case arriving: {
		const Coord goal (mirrored_ ?
		                  arrival_it_->XMirror() : *arrival_it_);
		const Coord delta (goal - pos_);
		double scale (delta.Length() / speed);
		if (scale < 1.0) {
			if (++arrival_it_ == arrival_.end()) stage_ = joining;
			NextPosition(speed);
			return;
		}
		if (scale == 1.0 && ++arrival_it_ == arrival_.end()) {
			stage_ = joining;
		}
		pos_ += delta / scale;
		break;
	}
	case joining: {
		const Coord cruise_pos (base_cruise_ + delta_cruise_);
		const Coord delta (cruise_pos - pos_);
		const double scale (delta.Length() / speed);
		if (scale <= 1.0) {
			stage_ = cruising;
			pos_ = cruise_pos;
		} else {
			pos_ += delta / scale;
		}
		break;
	}
	case cruising:
		pos_ = base_cruise_ + delta_cruise_;
		break;
	case attacking: {
		const Coord goal (*attack_it_);
		const Coord delta (goal - pos_);
		double scale (delta.Length() / (speed - 1));
		if (scale <= 1.0 && ++attack_it_ == attack_.end()) {
			stage_ = joining;
			pos_.y = -g_alien_height;
		} else {
			if (scale < 1.0) {
				NextPosition(speed);
				return;	
			}
			pos_ += delta / scale;
		}
		break;
	}
	} // switch
}


void Trajectory::BuildAttack()
{
	const int min_delta_y (RCoord::max / 64);
	attack_.clear();
	attack_.reserve(RCoord::max / min_delta_y);
	RCoord next (pos_.ToRCoord());
	do {
		next += RCoord(std::rand() % (RCoord::max/3) - RCoord::max/3/2,
		               std::rand() % (RCoord::max/4) + min_delta_y);
		attack_.push_back(next);
	} while (!next.YOut());
	attack_it_ = attack_.begin();
	stage_ = attacking;
}



/*
 *  Explosion
 */

Explosion::Explosion(Coord pos, Coord speed, XColor color, int duration)
: color_   (color)
, moves_   (0)
, duration_(duration)
{
	const int nb_particles (4 * duration * Details::Instance().Get());
	particles_.reserve(nb_particles);
	particles_speed_.reserve(nb_particles);
	for (int i (0); i < nb_particles; ++i) {
		using std::rand;
		const int rx (rand() % 16 + rand() % 16 - 16);
		const int ry (rand() % 16 + rand() % 16 - 16);
		const Coord particle (rx, ry);
		particles_.push_back(particle + pos);
		Coord particle_speed (particle / 4 + speed);
		particle_speed += Coord(rand() % 5 - 2, rand() % 5 - 2);
		particles_speed_.push_back(particle_speed);
	}
	XColor close_color = color_;
	X11::Inst().AllocColorAlways(&close_color);
	DrawAll(close_color.pixel);
}


void Explosion::Move()
{
	if (moves_ < duration_) {
		++moves_;
		DrawAll(X11::Inst().GetBlack());
		if (moves_ < duration_) {
			for (ParticlesCtr::size_type i (0);
			     i < particles_.size(); ++i) {
				particles_[i] += particles_speed_[i];
			}
			color_.red   = color_.red   * 98 / 100;
			color_.green = color_.green * 96 / 100;
			color_.blue  = color_.blue  * 96 / 100;
			if (color_.red < 48*1024 && color_.red > 36*1024) {
				color_.red   = color_.red   * 96 / 100;
				color_.green = color_.green * 92 / 100;
				color_.blue  = color_.blue  * 92 / 100;
			}
			XColor close_color = color_;
			X11::Inst().AllocColorAlways(&close_color);
			DrawAll(close_color.pixel);
		}
	}
}


void Explosion::DrawAll(X11::Color color) const
{
	X11::Inst().SetForeground(color);
	X11::Inst().DrawPoints(&particles_[0], particles_.size());
}



/*
 *  Projectile
 */

Projectile::Projectile(const Pix * pix, Coord pos, Coord speed)
: pix_  (pix)
, pos_  (pos)
, speed_(speed)
{
	if (!pix) throw std::runtime_error ("Null projectile pixmap");
	pix_->Draw(pos);
}


void Projectile::Move()
{
	pix_->Erase(pos_);
	pos_ += speed_;
	pix_->Draw(pos_);
}


bool Projectile::Collide(Coord pos2, Coord size) const
{
	// Aligned rectangles intersect <=> their projections intersects
	return !(pos_.x + pix_->Width() / 2  < pos2.x - size.x / 2 ||
	         pos_.x - pix_->Width() / 2  > pos2.x + size.x / 2 ||
		 pos_.y + pix_->Height() / 2 < pos2.y - size.y / 2 ||
		 pos_.y - pix_->Height() / 2 > pos2.y + size.y / 2);
}



/*
 *  Alien
 */

Alien::Alien(const Pix * pix, const Trajectory & trajectory, int speed)
: pix_        (pix)
, bomb_pix_   (PixKeeper::Instance().Get(bomb_xpm))
, trajectory_ (trajectory)
, speed_      (speed)
{
	if (!pix) throw std::runtime_error ("Null alien pix");
	pix_->Draw(Position());
}


void Alien::Move()
{
	pix_->Erase(Position());
	trajectory_.NextPosition(speed_);
	pix_->Draw(Position());
}


Bullet * Alien::Fire() const
{
	Coord fire_pos (Position());
	fire_pos.y += Height() / 2;
	const int bomb_speed_x (std::rand() % 3 + std::rand() % 3 - 2);
	const Coord bomb_speed (bomb_speed_x,
	                        speed_ * 3 / 2 - std::abs(bomb_speed_x));
	return new Bullet(bomb_pix_, fire_pos, bomb_speed);
}



/*
 *  Player
 */

Player::Player(int below_y)
: pix_            (PixKeeper::Instance().Get(player_xpm))
, bullet_pix_     (PixKeeper::Instance().Get(bullet_xpm))
, below_y_        (below_y)
, pos_            (X11::Inst().WindowWidth() / 2, PosY())
, shield_         (3)
, speed_          (2)
, fire_wait_      (0)
, fire_interval_  (24)
, multi_fire_     (1)
{
	pix_->Draw(pos_);
}


void Player::Move(int direction)
{
	if (fire_wait_ > 0) --fire_wait_;

	if (direction != 0) pix_->Erase(pos_);
	pos_.x += direction * speed_;
	if (pos_.x < Width() / 2) pos_.x = Width() / 2;
	else {
		const int max_right (X11::Inst().WindowWidth() - Width() / 2 - 1);
		if (pos_.x > max_right) pos_.x = max_right;
	}
	pos_.y = PosY();
	pix_->Draw(pos_);
}


void Player::Fire(std::list<Bullet*> * res) const
{
	for (int i (0); i < multi_fire_; ++i) {
		Coord speed (i - multi_fire_ / 2, -10);
		if (multi_fire_ % 2 == 0 && i >= multi_fire_ / 2) ++speed.x;
		Coord fire_pos (pos_);
		fire_pos.y -= Height() / 2;
		res->push_back(new Bullet(bullet_pix_, fire_pos, speed));
	}
}


std::list<Bullet*> Player::Fire()
{
	std::list<Bullet*> res;
	if (fire_wait_ <= 0) {
		fire_wait_ = fire_interval_;
		Fire(&res);
	}
	return res;
}


Explosion * Player::Collision(const Bullet * b)
{
	if (b->Collide(pos_, Coord(Width(), Height()))) {
		--shield_;
		XColor color;
		color.blue = 65535; color.red = color.green = 54*1024;
		return new Explosion(pos_, Coord(0,0), color, 50);
	}
	return 0;
}


bool Player::Collision(const Bonus * b)
{
	if (b->Collide(pos_, Coord(Width(), Height()))) {
		switch (b->Type()) {
		case Bonus::extra_speed:
			ExtraSpeed();
			break;
		case Bonus::extra_fire:
			ExtraFire();
			break;
		case Bonus::extra_shield:
			ExtraShield();
			break;
		}
		return true;
	}
	return false;
}


int Player::PosY() const
{
	return X11::Inst().WindowHeight() - below_y_ - Height() / 2;
}


void Player::ExtraFire()
{
	if (multi_fire_ < 9)
		if (fire_interval_ > 5) {
			fire_interval_ = fire_interval_ * 3 / 4;
			fire_wait_ = fire_wait_ * 3 / 4;
		} else {
			fire_interval_ = fire_interval_ *
			                 (multi_fire_+1) / multi_fire_;
			fire_wait_ = fire_wait_ *
			             (multi_fire_+1) / multi_fire_;
			++multi_fire_;
		}
}
